@@ -4,8 +4,6 @@ from dotenv import load_dotenv
 load_dotenv(override=False)
 
 PYTHON_ENV = os.getenv("PYTHON_ENV")
-REDIS_URL = os.getenv("REDIS_URL", "redis://localhost:6379")
-EXPERIMENT_ID = os.getenv("EXPERIMENT_ID")
 LOGGING_LEVEL = os.getenv("LOGGING_LEVEL", "INFO")
 
 DB_USER = (
@@ -32,5 +30,23 @@ DB_URI: str = (
 )
 OPENAI_API_KEY = os.getenv("OPENAI_API_KEY")
 LECRAPAUD_LOGFILE = os.getenv("LECRAPAUD_LOGFILE")
-LECRAPAUD_LOCAL = os.getenv("LECRAPAUD_LOCAL", False)
 LECRAPAUD_TABLE_PREFIX = os.getenv("LECRAPAUD_TABLE_PREFIX", "lecrapaud")
+LECRAPAUD_OPTIMIZATION_BACKEND = os.getenv(
+    "LECRAPAUD_OPTIMIZATION_BACKEND", "hyperopt"
+).lower()
+
+SENTRY_DSN = os.getenv("SENTRY_DSN")
+
+try:
+    SENTRY_TRACES_SAMPLE_RATE = float(os.getenv("SENTRY_TRACES_SAMPLE_RATE", "0"))
+except ValueError:
+    SENTRY_TRACES_SAMPLE_RATE = 0.0
+
+try:
+    SENTRY_PROFILES_SAMPLE_RATE = float(os.getenv("SENTRY_PROFILES_SAMPLE_RATE", "0"))
+except ValueError:
+    SENTRY_PROFILES_SAMPLE_RATE = 0.0
+
+LECRAPAUD_SAVE_FULL_TRAIN_DATA = (
+    os.getenv("LECRAPAUD_SAVE_FULL_TRAIN_DATA", "False").lower() == "true"
+)
