@@ -3,10 +3,11 @@
 from contextlib import contextmanager
 from sqlalchemy import create_engine, text
 from sqlalchemy.orm import sessionmaker
+from urllib.parse import urlparse
+
 from alembic.config import Config
 from alembic import command
 import os
-from urllib.parse import urlparse
 
 from lecrapaud.config import DB_USER, DB_PASSWORD, DB_HOST, DB_PORT, DB_NAME, DB_URI
 
@@ -27,6 +28,10 @@ else:
 
 def init_db(uri: str = None):
     global _engine, _SessionLocal, DATABASE_URL, DB_URI
+    if _SessionLocal is not None:
+        return
+
+    # Step 0: Set database URI
     if uri:
         if "mysql://" in uri and "pymysql://" not in uri:
             uri = uri.replace("mysql://", "mysql+pymysql://")
@@ -41,12 +46,10 @@ def init_db(uri: str = None):
         )
     print(f"Initializing database with URI: {DATABASE_URL}")
 
-    # Use urlparse for robust parsing
+    # Step 1: Create root engine without database
     parsed = urlparse(DATABASE_URL)
     db_name = parsed.path.lstrip("/")  # remove leading slash
     root_uri = f"{parsed.scheme}://{parsed.netloc}"
-
-    # Do not add trailing slash or database
     root_engine = create_engine(root_uri)
 
     # Step 2: Create database if it doesn't exist
@@ -73,6 +76,7 @@ def init_db(uri: str = None):
         autocommit=False,
         autoflush=False,
         bind=_engine,
+        expire_on_commit=False,  # Prevent detached instance errors
     )
 
     # Step 5: Apply Alembic migrations programmatically
